"""
Contains the DataFrame class
"""
import itertools
import csv
import functools
from .series import Series
from .indexers import ILocDF, LocDF
from .other_stuff import nan, is_bool, is_2d_bool


class DataFrame:
    """
    The only mutable attribute is the data.
    Shape is equal to the view.
    If a row is added, data is recreated and step is also updated.
    If a column is added, data is appended *only* if the dataframe view covers
    the entire dataset (shape equals len index, columns). Otherwise, a copy is made
    View is a tuple of two slices, for the row and column. Steps are not taken into
    account in view, it is high level. This is contrary to Series
    step = len(index) = shape(0) = view[0].stop - view[0].start
    len(columns) = shape(1) = view[1].stop - view[1].start
    """

    ITERABLE_1D = (list, set, tuple, Series)

    def __init__(self, data=None, index=None, columns=None):
        self.columns = tuple(columns) if columns else tuple()  # type: tuple
        self.index = tuple(index) if index else tuple()  # type: tuple
        self.data = []  # type: list
        self.name = None  # type: str
        self.step = 0  # type: int
        self.shape = (0, 0)  # type: tuple
        self.view = (slice(0, 0), slice(0, 0))  # type: tuple
        self.iloc = ILocDF(self)  # type: ILocDF
        self.loc = LocDF(self)  # type: LocDF

        if data is None:
            return
        if isinstance(data, dict):
            self.step = len(data[list(data.keys())[0]])
            self.data = list(itertools.chain(*data.values()))
            self.columns = tuple(data.keys())
        elif isinstance(data, list):
            if isinstance(data[0], self.ITERABLE_1D):
                # if they are series, try to extract the data
                try:
                    self.index = tuple(d.name for d in data)
                    self.columns = data[0].index
                except AttributeError:
                    pass
                self.step = len(data)

                data = list(zip(*data))
                for item in data:
                    self.data.extend(item)

            elif isinstance(data[0], dict):
                for d_dict in data:
                    key, val = next(iter(d_dict.items()))
                    self.columns = self.columns + (key,)
                    self.data.extend(val)
                self.step = len(val)

        if len(self.columns) == 0:
            self.columns = tuple(i for i in range(len(self.data) // self.step))

        if len(self.index) == 0:
            self.index = tuple(i for i in range(self.step))
        self.shape = (self.step, len(self.columns))
        self.view = (slice(0, self.shape[0]), slice(0, self.shape[1]))

    @classmethod
    def from_data(cls, data, index, columns, view, step):
        self = cls()
        self.data = data
        self.columns = columns
        self.index = index
        self.view = view
        self.step = step
        self.shape = (
            self.view[0].stop - self.view[0].start,
            self.view[1].stop - self.view[1].start,
        )
        self.iloc = ILocDF(self)
        self.loc = LocDF(self)

        return self

    @classmethod
    def class_init(cls, *args, **kwargs):
        return cls(*args, **kwargs)

    @classmethod
    def series_from_data(cls, *args):
        return Series.from_data(*args)

    def __str__(self):
        string = "DataFrame: " + "\n" + str(self.columns) + "\n"
        string += "\n".join(str(d) for d in zip(self.index, self.values))
        return string

    def __getitem__(self, cols):
        # gets here as slice and series
        if isinstance(cols, tuple):
            return self.loc[cols]
        elif isinstance(cols, slice) or is_bool(cols) or is_2d_bool(cols):
            return self.loc[cols, :]
        return self.loc[:, cols]

    def __setitem__(self, key, value):
        """
        This implementation just adds a new columns
        """
        # if two params were provided (i.e. if it's a tuple) forward it
        # if it is a slice or boolean, forward as row indexer
        # otherwise, forward as column indexer
        if isinstance(key, (tuple, self.__class__)):
            self.loc[key] = value
        elif isinstance(key, slice) or is_bool(key):
            self.loc[key, :] = value
        else:
            self.loc[:, key] = value

    def __delitem__(self, cols):
        self.drop(cols)

    def __lt__(self, other):
        df = self.copy()
        df.data = [item < other for item in df.data]
        return df

    def __le__(self, other):
        df = self.copy()
        df.data = [item <= other for item in df.data]
        return df

    def __gt__(self, other):
        df = self.copy()
        df.data = [item > other for item in df.data]
        return df

    def __ge__(self, other):
        df = self.copy()
        df.data = [item >= other for item in df.data]
        return df

    def __eq__(self, other):
        df = self.copy()
        df.data = [item == other for item in df.data]
        return df

    def __ne__(self, other):
        df = self.copy()
        df.data = [item != other for item in df.data]
        return df

    def __iter__(self):
        return iter(self.columns)

    def __invert__(self):
        df_cp = self.copy()
        for i in range(len(df_cp.data)):
            df_cp.data[i] = not df_cp.data[i]
        return df_cp

    def __len__(self):
        return self.shape[0]

    def __repr__(self):
        return str(self)

    def drop(self, labels=None):
        """
        Drop both removes a specified column and trims internal data, producing a copy.

        column_index: a column to drop
        :return:
        """
        to_delete = self.view[1].stop
        num = 0
        if isinstance(labels, str):
            to_delete = self.columns.index(labels)
            num = 1

        # build new dataset without the old columns
        data_cols = []

        for col_index in range(self.view[1].start, to_delete):
            data_cols.append(
                self.data[
                    self.view[0].start
                    + col_index * self.step : self.view[0].stop
                    + col_index * self.step
                ]
            )

        for col_index in range(to_delete + num, self.view[1].stop):
            data_cols.append(
                self.data[
                    self.view[0].start
                    + col_index * self.step : self.view[0].stop
                    + col_index * self.step
                ]
            )
        self.data = []
        for col in data_cols:
            self.data.extend(col)

        #    and adjust our indexing
        self.columns = self.columns[0:to_delete] + self.columns[to_delete + num :]
        self.shape = (self.shape[0], self.shape[1] - num)
        self.view = (slice(0, len(self.index)), slice(0, len(self.columns)))
        self.step = len(self.index)

    def copy(self):
        """
        Creates a copy of the dataframe and trims the data with self.drop
        :return:
        """
        df = DataFrame.from_data(
            self.data, self.index, self.columns, self.view, self.step
        )
        df.drop()
        return df

    def equals(self, other):
        return (self.values == other.values) and (self.shape == other.shape)

    @property
    def values(self):
        data_rows = []
        for row_index in range(self.view[0].start, self.view[0].stop):
            data_rows.append(
                self.data[
                    row_index
                    + self.step * self.view[1].start : row_index
                    + self.step * self.view[1].stop : self.step
                ]
            )
        return data_rows

    def bound_int_to_df(self, raw_int, axis):
        """
        Transforms an index int to the actual axis index of data,
        taking bounds into account
        e.g.
        [0,| 1, 2, 3, 4, 5, | 6]
        If `2` is given, it should access "3", thus index 2 is actually 3.
        -1 becomes 5
        6 would be an index error
        Slices are handled by a bound_slice_to_df

        :param raw_int:
        :param axis:
        :return:
        """
        if axis in [0, "row", "rows"]:
            view_min = self.view[0].start
            view_max = self.view[0].stop
        elif axis in [1, "column", "columns"]:
            view_min = self.view[1].start
            view_max = self.view[1].stop
        else:
            raise UserWarning

        # handle negative ints
        if raw_int < 0:
            start = view_max + raw_int
        else:
            start = view_min + raw_int

        # check bounds
        if start > view_max or start < view_min:
            raise IndexError

        return start

    def bound_slice_to_df(self, raw_slice, axis):
        """
        Transforms a slice to the actual axis view for the data
        :param raw_slice: relative slice
        :return: slice to underlying data
        """
        if axis in [0, "row", "rows"]:
            view_start = self.view[0].start
            view_stop = self.view[0].stop
        elif axis in [1, "column", "columns"]:
            view_start = self.view[1].start
            view_stop = self.view[1].stop
        else:
            pass

        if raw_slice.start:
            if raw_slice.start < 0:
                start = max(view_stop + raw_slice.start, view_start)
            else:
                start = min(view_start + raw_slice.start, view_stop)
        else:
            start = view_start

        if raw_slice.stop:
            if raw_slice.stop < 0:
                stop = max(view_stop + raw_slice.stop, view_start)
            else:
                stop = min(view_start + raw_slice.stop, view_stop)
        else:
            stop = view_stop
        return slice(start, stop)

    def bound_iterable_to_df(self, raw_iter, axis):
        """
        Converts indicies to the actual data indicies
        :param raw_iter:
        :param axis:
        :return:
        """
        return [self.bound_int_to_df(item, axis) for item in raw_iter]

    def convert_slice(self, raw_slice, axis):
        """
        Removes the None from the slice and replaces it with length of rows or columns.
        doesn't adjust based on view
        :param raw_slice:
        :param axis:
        :return:
        """
        if axis in [0, "row", "rows"]:
            max_stop = len(self.index)
        elif axis in [1, "columns", "cols", "col"]:
            max_stop = len(self.columns)

        if not raw_slice.start:
            start = 0
        elif raw_slice.start < 0:
            start = max(0, max_stop + raw_slice.start)
        else:
            start = raw_slice.start

        if not raw_slice.stop:
            stop = max_stop
        elif raw_slice.stop < 0:
            stop = max(0, max_stop + raw_slice.stop)
        else:
            stop = raw_slice.stop
        return slice(start, stop)

    def is_view(self):
        """
        Determines whether or not the dataframe is a view of another dataframe.
        Checks if the shape is the sshape of the entire data
        :return:
        """
        return self.shape[0] != self.step or self.shape[1] != len(self.data) / self.step

    def index_of(self, item, axis=0):
        """
        Returns the integer index of a column/index label
        :param item: iterable or label
        :param axis: 0 - search index; 1 - search column labels
        :return: list(int) or int
        """
        if axis in [0, "rows", "row"]:
            names = self.index
        else:
            names = self.columns
        if isinstance(item, self.ITERABLE_1D):
            # bypass for boolean
            if is_bool(item):
                return item
            return [names.index(i) for i in item]
        elif isinstance(item, slice):
            start = None if item.start is None else names.index(item.start)
            stop = None if item.stop is None else names.index(item.stop)
            return slice(start, stop)
        else:
            try:
                return names.index(item)
            except ValueError:
                return None

    def add_empty_series(self, name, axis=0):
        """
        Adds a new row/column to a dataframe.

        If the dataframe is a view, it will make a copy of itself and trim its
        :param name: Index/column name
        :param axis: int; 0 - adds a row. 1 - adds a column
        :return: None. Does so in place.
        """
        # if its a view, make a copy
        if self.is_view():
            self.drop()

        # Add a row
        if axis == 0:
            self.index = self.index + (name,)
            ndata = []
            for i in range(self.shape[1]):
                ndata.extend(self.data[i * self.step : (i + 1) * self.step] + [nan])
            self.data = ndata
            self.shape = (self.shape[0] + 1, self.shape[1])
            self.view = (slice(self.view[0].start, self.view[0].stop + 1), self.view[1])
            self.step = self.step + 1
        # add a column
        elif axis == 1:
            self.columns = self.columns + (name,)
            self.data = self.data + [nan] * self.shape[0]
            self.shape = (self.shape[0], self.shape[1] + 1)
            self.view = (self.view[0], slice(self.view[1].start, self.view[1].stop + 1))

    def append(self, other, ignore_index=False):
        """
        Appends either a DataFrame or Series.

        :param other: DataFrame or Series
        :param ignore_index: Bool, If false, will create a new index
        :return: DataFrame
        """

        if isinstance(other, self.__class__):
            # Determine new columns
            columns = self.columns + tuple(
                col for col in other.columns if col not in self.columns
            )
            # Create data, with nans if there are new columns
            data_columns = []
            for col in columns:
                if col in self.columns:
                    temp = self[col].values
                else:
                    temp = [nan] * len(self)
                if col in other.columns:
                    temp += other[col].values
                else:
                    temp += [nan] * len(other)
                data_columns.append(temp)

            # new index
            if ignore_index:
                index = None
            else:
                index = self.index + other.index
        elif isinstance(other, self.ITERABLE_1D):
            if not ignore_index and other.name is None:
                raise TypeError(
                    "Can only append a Series/Array if ignore_index=True or if the Series has a name"
                )
            columns = self.columns + tuple(
                col for col in other.index if col not in self.columns
            )
            # generate new data
            data_columns = []
            for col in columns:
                # df data
                if col in self.columns:
                    temp = self[col].values
                else:
                    temp = [nan] * len(self)
                # iterable data
                if col in other.index:
                    temp.append(other[col])
                else:
                    temp.append(nan)
                data_columns.append(temp)

            # new index
            if ignore_index:
                index = None
            else:
                index = self.index + other.name
        # new dataframe
        return self.class_init(
            {k: v for k, v in zip(columns, data_columns)}, index=index
        )

    def applymap(self, func):
        cp = self.copy()
        cp.data = [func(item) for item in cp.data]
        return cp

    def apply(self, func, axis=0, dropna=True):

        res = []
        index = []
        if axis == 0:
            iterator = self.itercols
        else:
            iterator = self.iterrows

        for item in iterator():
            item = list(item)
            try:
                # if the function is reducing and works on an iterable, try that
                if dropna:
                    item[1] = item[1].dropna()
                # it it was all nans, just put a nan there
                if len(item[1]) == 0:
                    res.append(nan)
                else:
                    res.append(func(item[1]))
            except TypeError:
                try:
                    # otherwise, elementwise
                    res.append(item[1].apply(func))
                except TypeError:
                    # otherwise, skip
                    continue

            index.append(item[0])
        if isinstance(res[0], Series):
            if axis == 0:
                return self.class_init(res).transpose()
            else:
                return self.class_init(res)
        else:
            return Series(res, index, name=self.name)

    def iterrows(self):
        for i in range(len(self)):
            try:
                yield (self.index[i], self.iloc[i, :])
            except:
                return

    def itercols(self):
        for i in range(len(self.columns)):
            try:
                yield (self.columns[i], self.iloc[:, i])
            except:
                return

    def iteritems(self):
        return self.itercols()

    def transpose(self):
        new_cols = self.index
        new_index = self.columns
        data = list(zip(*self.values))
        cp = self.class_init(data, columns=new_cols, index=new_index)
        return cp

    def sort_values(self, by, ascending=True, axis=0, na_position="last"):
        if axis == 0:
            it = self.itercols
            ser = self.loc[:, by]

        else:
            it = self.iterrows
            ser = self.loc[by, :]

        new_index = ser.sort_values(ascending, na_position).index
        res = []
        cols = []

        for ser in it():
            res.append(ser[1].loc[new_index])
            cols.append(ser[0])

        cp = self.class_init(res, columns=cols)
        if axis == 0:
            cp = cp.transpose()
        return cp

    def reset_index(self, drop=False):
        cp = self.copy()
        if not drop:
            cp["index"] = cp.index
            cp = cp.loc[:, cp.columns[-1:] + cp.columns[:-1]]
        cp.index = tuple(i for i in range(len(self)))

        return cp

    def groupby(self, by):
        """
        Simple groupby implementation
        """
        gb = GroupBy()

        for item in self[by].unique():
            df = self.loc[self[by] == item, :]
            df.name = item
            df.drop(by)
            gb.dfs.append(df)
        return gb

    def mean(self, axis=0, dropna=True):
        return self.apply(lambda x: sum(x) / len(x), axis=axis, dropna=dropna)

    def sum(self, axis=0, dropna=True):
        if axis == 0:
            iterator = self.iterrows
        else:
            iterator = self.itercols

        res = None
        for item in iterator():
            if item is None:
                res = item[1]
            else:
                res += item[1]
        res.name = self.name
        return res


class GroupBy:
    """
    GroupBy class for DataFrame
    """

    def __init__(self):
        """
        Collection of DataFrames
        """
        self.dfs = []

    def apply(self, func, axis=0, dropna=True):
        """
        Applies a method
        """
        res_ser = []
        for df in self.dfs:
            res_ser.append(df.apply(func, axis=axis, dropna=dropna))
        return DataFrame(res_ser)

    def __sum__(self):
        pass

    def __getattr__(self, item):
        """
        Returns a method containing a for loop of partialized
        methods, awaiting *args and **kwargs
        """
        return functools.partial(self.loop_func, method_name=item)

    def loop_func(self, method_name, *args, **kwargs):
        """
        The function to be executed with args and kwargs
        """
        res_ser = []
        for df in self.dfs:
            res_ser.append(df.__getattribute__(method_name)(*args, **kwargs))
        return DataFrame(res_ser)


def read_csv(filepath, sep=",", header=0, names=None, index_col=None):
    """
    Reads CSV file into dataframe
    """
    # see if there is an fspath
    try:
        filepath = filepath.__fspath__()
    except:
        pass
    index = []
    columns = []
    data = []
    if names is not None:
        header = None
        columns = names
    with open(filepath) as csvfile:
        spamreader = csv.reader(csvfile, delimiter=sep)
        for i, row in enumerate(spamreader):
            if isinstance(header, int) and header == i:
                if isinstance(index_col, int):
                    columns = row[:index_col] + row[index_col + 1 :]
                else:
                    columns = row
                continue

            if isinstance(index_col, int):
                index.append(row[index_col])
                data.append(row[:index_col] + row[index_col + 1 :])
            else:
                data.append(row)

    return DataFrame(data, columns=columns, index=index)



from functools import reduce
from copy import copy


class NaN:
    def __eq__(self, *args):
        return False

    def __lt__(self, *args):
        return False

    def __le__(self, *args):
        return False

    def __gt__(self, *args):
        return False

    def __ge__(self, *args):
        return False

    def __str__(self):
        return "NaN"

    def __repr__(self):
        return self.__str__()

    def __add__(self, other):
        return self

    def __radd__(self, other):
        return self

    def __truediv__(self, other):
        return self

    def __floordiv__(self, other):
        return self

    def __mul__(self, other):
        return self

    def __pow__(self, other):
        return self

    def __mod__(self, other):
        return self


nan = NaN()


def is_bool(key):
    """
    Checks if the first value in some kind of item is a boolean value
    """
    try:
        item0 = key.iloc[0]
    except AttributeError:
        if isinstance(key, (list, tuple, set)):
            item0 = key[0]
        else:
            item0 = key
    if isinstance(item0, bool):
        return True
    return False


def is_2d_bool(key):
    """
    Checks if an object is a 2D bool key
    """
    try:
        item0 = key.iloc[0, 0]
    except AttributeError:
        try:
            item0 = key[0][0]
        except:
            return False
    if isinstance(item0, bool):
        return True
    return False


def invert(item):
    """
    Copies and inverts a list or nested list
    :param item: a nd iterable of boolean values
    :return: ~item
    """
    res = []
    if hasattr(item[0], "__len__"):
        for i in range(len(item)):
            res.append(invert(item[i]))
    else:
        res = [not val for val in item]
        return res
    return res


def concat(items, axis=0, join="outer", ignore_index=False):
    """
    Concatenates DataFrames or Series

    :param items: list, Series or DataFrames
    :param axis: int, default 0
    :param join: str, 'inner' or 'outer'
    :param ignore_index: bool, default False
    :return: DataFrame
    """
    if hasattr(items[0], "columns"):
        return concat_df(items, axis, join, ignore_index)
    else:
        return concat_ser(items, axis, join, ignore_index)


def concat_df(items, axis=0, join="outer", ignore_index=False):
    """
    Concatenates two or more dataframes

    :param other: DataFrame or Series
    :param ignore_index: Bool, If false, will create a new index
    :return: DataFrame
    """
    # append top and bottom
    if axis == 0:
        join_on = "columns"
        index_on = "index"
    else:
        join_on = "index"
        index_on = "columns"

    # build columns
    indices = [getattr(item, join_on) for item in items]
    if join == "outer":
        indices = list(reduce(lambda x, y: list_union(x, y), indices))
    else:
        indices = list(reduce(lambda x, y: list_intersection(x, y), indices))

    # Create data, with nans if there are new columns
    data = []
    for idx in indices:
        temp = []
        for item in items:
            if idx in getattr(item, join_on):
                if axis == 0:
                    temp += item[idx].values
                else:
                    temp += item.loc[idx, :].values
            else:
                if axis == 0:
                    length = item.shape[0]
                else:
                    length = item.shape[1]
                temp += [nan] * length
        data.append(temp)

    # new index. index if axis=0, columns if axis=1
    if ignore_index:
        index = None
    else:
        index = []
        for item in items:
            index += getattr(item, index_on)

    if axis == 0:
        return items[0].class_init({k: v for k, v in zip(indices, data)}, index=index)
    return items[0].class_init(data, columns=index, index=indices)


def concat_ser(items, axis=0, join="outer", ignore_index=False):
    """
    Concatenates tweo or more Series

    :param items: list, Series
    :param axis: int, default 0
    :param join: str, 'inner' or 'outer'
    :param ignore_index: bool, default False
    :return: DataFrame
    """
    # ugly, but necessary :/
    from .dataframe import DataFrame

    # axis is zero, just append series to self
    if axis == 0:
        data = []
        index = []
        for item in items:
            data += item.values
            index += item.index
        return items[0].__class__(data, index=index, name=items[0].name)

    # otherwise...
    # generate new columns or index  based on join type
    new_index = [item.index for item in items]
    if join == "outer":
        new_index = list(reduce(lambda x, y: list_union(x, y), new_index))
    else:
        new_index = list(reduce(lambda x, y: list_intersection(x, y), new_index))

    # make the data
    data = []
    for item in items:
        data.append([item[idx] if idx in item.index else nan for idx in new_index])

    # make the index
    if ignore_index:
        columns = None
    else:
        columns = [item.name for item in items]
    return DataFrame(list(zip(*data)), columns=columns, index=new_index)


def list_intersection(a, b):
    """
    Returns the intersection of two lists
    """
    return [item for item in a if item in b]


def list_union(a, b):
    """
    Returns a deduped union of two lists
    """
    c = list(copy(a))
    for item in b:
        if item not in a:
            c.append(item)
    return c



"""
Contains a DataFrame and Series class, as well as a nan type
"""
__version__ = "0.0.1"
from .dataframe import DataFrame, read_csv
from .series import Series
from .other_stuff import nan, concat

# def clean_slices(phase, info):
#     if phase == "stop":
#         return
#     obj_dict = {}
#     data_dict = {}
#     for obj in gc.get_objects():
#         if isinstance(obj, (Series, DataFrame)):
#             d_id = id(obj.data)
#             data_dict[d_id] = obj.data
#             try:
#                 obj_dict[d_id].append(obj)
#             except:
#                 obj_dict[d_id] = [obj]
#
#     # check if the series is part of a dataframe. if so, skip it.
#
#     for d_id, objs in obj_dict.items():
#         # find minimum start or maximum stop
#         start = objs[0].view.start
#         stop = objs[0].view.stop
#         for obj in objs[1:]:
#             # early stopping if both are None. Data is still in use somewhere
#             if start is None and stop is None:
#                 break
#             if (start is not None) and (
#                 (obj.view.start is None) or obj.view.start < start
#             ):
#                 start = obj.view.start
#             if stop is not None and ((obj.view.stop is None) or obj.view.stop > stop):
#                 stop = obj.view.stop
#         # if both are None, don't do anything. No trimming necessary
#         if start is None and stop is None:
#             continue
#         if start is None:
#             start = 0
#         # adjust the data and view for each by trimming data and subtracting the start
#         data_dict[d_id][:] = data_dict[d_id][start:stop]
#         for obj in objs:
#             new_start = new_stop = None
#             if obj.view.start is not None:
#                 new_start = obj.view.start - start
#             if obj.view.stop is not None:
#                 new_stop = obj.view.stop - start
#             obj.view = slice(new_start, new_stop)
#             obj.index[:] = obj.index[start:stop]



"""
Contains the Series class
"""
from datetime import datetime
from .indexers import LocSer, ILocSer
from .other_stuff import nan, is_bool
import functools


class Series:
    """
    view: the actual view of the data, including step
    """

    ITERABLE_1D = (list, set, tuple)

    @classmethod
    def from_data(cls, data, index, name=None, view=slice(None, None)):
        """
        Creates a Series from data and an index
        """
        self = cls()
        self.data = data  # full 1D dataset.
        self.index = tuple(index)  # index, unique to series
        self.name = name
        self.view = view  # data[view] = the values
        self.iloc = ILocSer(self)
        self.loc = LocSer(self)
        return self

    def __init__(self, data=None, index=None, name=None):
        view = None
        if isinstance(data, self.__class__):
            data = data.data
            index = data.index
            name = data.name
            view = data.view
        elif isinstance(data, (list, set, tuple)):
            data = list(data)
            view = slice(0, len(data), 1)
        elif isinstance(data, dict):
            name, data = next(iter(data.items()))
            view = slice(0, len(data), 1)

        if data and index is None:
            index = tuple(range(len(data)))
        self.data = data
        self.view = view
        self.index = tuple(index) if index else None
        self.name = name
        self.iloc = ILocSer(self)
        self.loc = LocSer(self)
        self.str = STR(self)
        self.dt = DT(self)

    def __setitem__(self, key, value):
        self.loc.__setitem__(key, value)

    def __getitem__(self, item):
        return self.loc.__getitem__(item)

    def __str__(self):
        return (
            "Series Name: "
            + str(self.name)
            + "\n"
            + "\n".join(
                "%s: %s" % (item[0], item[1])
                for item in zip(self.index, self.data[self.view])
            )
        )

    def __repr__(self):
        return str(self)

    def index_of(self, item):
        """
        Returns the integer index of values in the index.
        If it is a tuple and it's not found, raises KeyError.
        If it is another kind of iterable and not found, adds None

        :param item: slice, iterable, any hashable object

        :return: list, slice, int
        """
        names = self.index

        if isinstance(item, self.ITERABLE_1D + (self.__class__,)):

            items = []
            for i in item:
                try:
                    items.append(names.index(i))
                except ValueError:
                    if isinstance(item, tuple):
                        raise KeyError("%s not found in index." % i)
                    else:
                        items.append(None)
            return items
        elif isinstance(item, slice):
            try:
                start = None if item.start is None else names.index(item.start)
                stop = None if item.stop is None else names.index(item.stop)
            except ValueError:
                raise KeyError(
                    "At least one of the following values is not in the index: %s %s"
                    % (item.start, item.stop)
                )
            return slice(start, stop)
        else:
            try:
                return names.index(item)
            except ValueError:
                return None

    @property
    def values(self):
        return self.data[self.view]

    def __lt__(self, other):
        ser = self.copy()
        ser.data = [item < other for item in ser.data]
        return ser

    def __le__(self, other):
        ser = self.copy()
        ser.data = [item <= other for item in ser.data]
        return ser

    def __gt__(self, other):
        ser = self.copy()
        ser.data = [item > other for item in ser.data]
        return ser

    def __ge__(self, other):
        ser = self.copy()
        ser.data = [item >= other for item in ser.data]
        return ser

    def __eq__(self, other):
        if isinstance(other, (self.ITERABLE_1D, type(self))):
            if isinstance(other, type(self)):
                if other.index != self.index:
                    raise ValueError(
                        "Can only compare identically-labeled Series objects"
                    )
                else:
                    data = [item == o for item, o in zip(self, other)]
            else:
                data = [item == o for item, o in zip(self, other)]
        else:
            data = [item == other for item in self]

        ser = self.copy()
        ser.data = data
        return ser

    def __ne__(self, other):
        ser = self.copy()
        ser.data = [item != other for item in ser.data]
        return ser

    def drop(self, labels=None):
        """
        Trims the series, breaking any shared data with others

        column_index: a column to drop
        :return:
        """
        to_delete = self.view.stop
        if labels in self.index:
            to_delete = self.index.index(labels)

        self.data = (
            self.data[self.view][0:to_delete] + self.data[self.view][to_delete + 1 :]
        )
        self.index = self.index[0:to_delete] + self.index[1 + to_delete :]

        #    and adjust our indexing
        self.view = slice(0, len(self.index), 1)
        return self

    def copy(self):
        ser = self.from_data(
            self.data[self.view], self.index, self.name, slice(0, len(self.index), 1)
        )
        return ser

    def extend(self, index_name, value=None, num=1):
        self.drop()
        self.data.extend([nan] * num)
        if isinstance(index_name, self.ITERABLE_1D + (self.__class__,)):
            self.index = self.index + tuple(index_name)
        else:
            self.index = self.index + (index_name,)
        self.view = slice(self.view.start, self.view.stop + num, 1)

    def __len__(self):
        return len(self.index)

    def __next__(self):
        for i in range(len(self)):
            try:
                yield self.iloc[i]
            except IndexError:
                return

    def __iter__(self):
        for val in self.values:
            yield val

    def bound_slice(self, slc):
        """
        Converts a slice to the actual slice used to reference the
        data. Does not raise bounds error.
        """
        start = slc.start
        stop = slc.stop

        if start is None:
            start = 0
        if stop is None:
            stop = len(self)
        if start < 0:
            start = max(self.view.stop + start * self.view.step, self.view.start)
        else:
            start = min(self.view.start + start * self.view.step, self.view.stop)
        if stop < 0:
            stop = max(self.view.stop + stop * self.view.step, self.view.start)
        else:
            stop = min(self.view.start + stop * self.view.step, self.view.stop)

        return slice(start, stop, self.view.step)

    def bound_int(self, idx):
        """
        Converts an index to the actual index of the underlying
        data. Does not raise out of range errors.
        :param idx: int, desired index.values data
        :return: int, actual index of data
        """
        if idx < 0:
            idx = len(self) + idx
        idx = self.view.start + idx * self.view.step
        return idx

    def bound_iterable(self, iterable):
        """
        Converts an iterable of desired indecies to actual index numbers
        :param iterable:
        :return: list, a lit of index of the underlying data
        """
        return [self.bound_int(item) for item in iterable]

    def astype(self, type_name, copy=True):
        res = self.apply(type_name)
        if copy:
            return res
        self.iloc[:] = res.values

    def apply(self, func, *args, **kwargs):
        cp = self.copy()
        for i, val in enumerate(cp.values):
            cp.iloc[i] = func(val, *args, **kwargs)
        return cp

    def sort_values(self, ascending=True, na_position="last"):
        """
        sort_values sorts a series using the python built-in sorted function.
        :param ascending: bool, whether or not sorted values should be ascending
        :param na_position: str, 'first' or 'last'
        :return: Series, sorted
        """
        # remove nans
        indices = [i for i, x in enumerate(self.values) if x is nan]
        nan_index = [self.index[i] for i in indices]
        new_values = self.values
        new_index = list(self.index)
        for i, idx in enumerate(indices):
            del new_values[idx - i]
            del new_index[idx - i]

        reverse = not ascending
        new_values, new_index = zip(
            *sorted(zip(new_values, new_index), reverse=reverse)
        )

        if na_position == "last":
            new_values = list(new_values) + [nan] * len(nan_index)
            new_index = list(new_index) + list(nan_index)
        else:
            new_values = [nan] * len(nan_index) + list(new_values)
            new_index = list(nan_index) + list(new_index)

        return self.from_data(new_values, new_index, name=self.name)

    def unique(self):
        return list(set(self.values))

    def __add__(self, other):
        cp = self.copy()
        if isinstance(other, self.ITERABLE_1D + (self.__class__,)):
            for i, val in enumerate(other):
                cp.data[i] += val
        else:
            for i in range(len(self)):
                cp.data[i] += other
        return cp

    def __radd__(self, other):
        try:
            return self + other
        except:
            return self
        # if other is not 0:
        #     return self + other
        # else:
        #     return self

    def __mul__(self, other):
        cp = self.copy()
        if isinstance(other, self.ITERABLE_1D + (self.__class__,)):
            for i, val in enumerate(other):
                cp.data[i] *= val
        else:
            for i in range(len(self)):
                cp.data[i] *= other
        return cp

    def __truediv__(self, other):
        cp = self.copy()
        if isinstance(other, self.ITERABLE_1D + (self.__class__,)):
            for i, val in enumerate(other):
                cp.data[i] /= val
        else:
            for i in range(len(self)):
                cp.data[i] /= other
        return cp

    def __floordiv__(self, other):
        cp = self.copy()
        if isinstance(other, self.ITERABLE_1D + (self.__class__,)):
            for i, val in enumerate(other):
                cp.data[i] //= val
        else:
            for i in range(len(self)):
                cp.data[i] //= other
        return cp

    def __pow__(self, other):
        cp = self.copy()
        if isinstance(other, self.ITERABLE_1D + (self.__class__,)):
            for i, val in enumerate(other):
                cp.data[i] **= val
        else:
            for i in range(len(self)):
                cp.data[i] **= other
        return cp

    def dropna(self):
        cp = self.copy()
        cp = cp[~self.isna()]
        return cp

    def isna(self):
        """
        Returns a bool of whether or not an item is a nan
        :return: Series
        """
        cp = self.copy()
        cp.data = [item is nan for item in self.values]
        return cp

    def __invert__(self):
        cp = self.copy()
        for i in range(len(cp)):
            cp.data[i] = not cp.data[i]
        return cp

    def mean(self, dropna=True):
        if dropna:
            values = self.dropna().values
        return sum(values) / len(values)

    def sum(self, dropna=True):
        if dropna:
            values = self.dropna().values
        res = values[0]
        for item in values[1:]:
            res += item
        return res


class STR:
    def __init__(self, obj):
        self.obj = obj

    def __getattr__(self, item):
        # must return a partialed apply, which can accept args
        return functools.partial(self.obj.apply, func=getattr(str, item))


class DT:
    def __init__(self, obj):
        self.obj = obj

    def __getattr__(self, item):
        # must return a partialed apply, which can accept args
        return functools.partial(self.obj.apply, func=getattr(datetime, item))



"""
Setup for the module
"""
from setuptools import setup
import codecs
import os.path


def read(rel_path):
    """
    Opens a path file
    """
    here = os.path.abspath(os.path.dirname(__file__))
    with codecs.open(os.path.join(here, rel_path), "r") as fp:
        return fp.read()


def get_version(rel_path):
    """
    Pulls version from module
    """
    for line in read(rel_path).splitlines():
        if line.startswith("__version__"):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    else:
        raise RuntimeError("Unable to find version string.")


setup(
    name="pambdas",
    version=get_version("pambdas/__init__.py"),
    description="A pure-Python DataFrame implementation, with a similar API to Pandas.",
    author="Daniel Hitchcock",
    packages=["pambdas"],
)



"""
Contains the .loc and .iloc indexers for both DataFrames and Series
"""
from copy import copy
import itertools

from .other_stuff import is_bool, is_2d_bool, nan, invert


class ILocDF:
    """
    ILoc indexer for dataframes
    """

    def __init__(self, obj=None):
        """
        Initializes the indexer
        :param obj: Series
        """
        if obj is None:
            return
        self.obj = obj

    def __getitem__(self, items):
        """
        Getitem for DataFrames based on index number
        """
        data = self.obj.data
        index = self.obj.index
        columns = self.obj.columns
        step = self.obj.step
        view = self.obj.view
        name = None
        # if it's a tuple, its multiple indicies. Otherwise, its one item, so
        # make a dummy index
        if isinstance(items, tuple):
            items = list(items)
        else:
            items = [items, slice(None, None)]

        if is_2d_bool(items[0]):
            if isinstance(items[0], self.obj.__class__):
                items[0] = items[0].values

            df_cp = self.obj.copy()
            df_cp[invert(items[0])] = nan
            return df_cp
        data_items = copy(items)

        # convert to bool, or bound
        for i, item in enumerate(items):
            if isinstance(item, self.obj.ITERABLE_1D):
                # if it's a boolean
                if is_bool(item):
                    items[i] = [i for i, val in enumerate(item) if val]
                data_items[i] = self.obj.bound_iterable_to_df(items[i], axis=i)

            elif isinstance(item, slice):
                items[i] = self.obj.convert_slice(item, axis=i)
                data_items[i] = self.obj.bound_slice_to_df(items[i], axis=i)
            elif isinstance(item, int):
                data_items[i] = self.obj.bound_int_to_df(item, axis=i)

        #################
        # Returns an item
        #################
        if isinstance(items[0], int) and isinstance(items[1], int):
            # eg [1, 0]
            return data[data_items[0] + step * data_items[1]]
        ##################
        # Returns a Series
        ##################
        if isinstance(items[0], slice) and isinstance(items[1], int):
            # eg [1:3, 0]
            index = index[items[0]]
            name = columns[items[1]]
            view = slice(
                data_items[0].start + data_items[1] * step,
                data_items[0].stop + data_items[1] * step,
                1,
            )
        elif isinstance(items[0], int) and isinstance(items[1], slice):
            # eg .iloc[0, 1:3]
            name = index[items[0]]
            index = columns[items[1]]
            start = data_items[0] + step * data_items[1].start
            stop = data_items[0] + step * data_items[1].stop
            view = slice(start, stop, step)
        elif isinstance(items[0], int) and isinstance(items[1], self.obj.ITERABLE_1D):
            # eg .iloc[0, [1, 2, 3]]
            name = index[items[0]]
            index = tuple(columns[i] for i in items[1])
            data = [data[data_items[0] + step * i] for i in data_items[1]]
            # returns a copy of the data, so index starts at zero
            view = slice(0, len(items[1]))
        elif isinstance(items[0], self.obj.ITERABLE_1D) and isinstance(items[1], int):
            # eg .iloc[[1, 2, 3], 0]
            name = columns[items[1]]
            index = tuple(index[i] for i in items[0])
            data = [data[i + step * data_items[i][1]] for i in data_items[i][0]]
            view = slice(0, len(items[0]))

        #####################
        # Returns a DataFrame
        #####################
        elif isinstance(items[0], slice) and isinstance(items[1], slice):
            # e.g. .iloc[1:3, :]
            name = columns[items[1]]
            index = index[items[0]]
            view = tuple(data_items)
        elif isinstance(items[0], self.obj.ITERABLE_1D) and isinstance(items[1], slice):
            # e.g. .iloc[[1, 2], :]
            # iterate through row
            ndata = []
            for col_index in range(data_items[1].start, data_items[1].stop):
                ndata.extend([data[i + col_index * step] for i in data_items[0]])
            data = ndata
            name = columns[items[1]]
            index = tuple(index[i] for i in items[0])
            step = len(index)
            # retuns a copy, so view starts at zero
            view = (slice(0, step), slice(0, len(name)))
        elif isinstance(items[0], slice) and isinstance(items[1], self.obj.ITERABLE_1D):
            # e.g. .iloc[:, [1,2]
            ndata = []
            for i in data_items[1]:
                ndata.extend(
                    data[data_items[0].start + i * step : data_items[0].stop + i * step]
                )
            data = ndata
            index = index[items[0]]
            name = tuple(columns[i] for i in items[1])
            step = len(index)
            # return a copy, view starts at zero
            view = (slice(0, step), slice(0, len(name)))
        elif isinstance(items[0], self.obj.ITERABLE_1D) and isinstance(
            items[1], self.obj.ITERABLE_1D
        ):
            # e.g. .iloc[:, [1,2]
            ndata = []
            for i in data_items[1]:
                ndata.extend([data[j + i * step] for j in data_items[0]])
            data = ndata
            index = tuple(index[i] for i in items[0])
            name = tuple(columns[i] for i in items[1])
            step = len(index)
            # return a copy, view starts at zero
            view = (slice(0, step), slice(0, len(name)))

        if isinstance(index, tuple) and isinstance(name, (str, int)):
            return self.obj.series_from_data(data, index, name, view)
        if isinstance(index, tuple) and isinstance(name, tuple):
            return self.obj.from_data(data, index, name, view, step)
        raise IndexError(
            "Unhandled params in DF .iloc getitem. Perhaps your are not referencing by index"
        )

    def __setitem__(self, items, value):
        """
        Setitem for DataFrames based on index number
        """
        data = self.obj.data
        step = self.obj.step

        # if it's a tuple, its multiple indicies. Otherwise, make a dummy index
        if isinstance(items, tuple):
            items = list(items)
        else:
            items = [items, slice(None, None)]
        data_items = copy(items)

        # convert to bool, or bound

        for i, item in enumerate(items):
            if is_2d_bool(item):
                pass
            elif isinstance(item, self.obj.ITERABLE_1D):
                # if it's a boolean
                if is_bool(item):
                    items[i] = [i for i, val in enumerate(item) if val]
                data_items[i] = self.obj.bound_iterable_to_df(items[i], axis=i)
            elif isinstance(item, slice):
                items[i] = self.obj.convert_slice(item, axis=i)
                data_items[i] = self.obj.bound_slice_to_df(items[i], axis=i)
            elif isinstance(item, int):
                data_items[i] = self.obj.bound_int_to_df(item, axis=i)

        del items

        #################
        # Sets an item
        #################
        if isinstance(data_items[0], int) and isinstance(data_items[1], int):
            # eg [1, 0]
            data[data_items[0] + step * data_items[1]] = value
        ##################
        # Sets a 1D section
        ##################
        if isinstance(data_items[0], slice) and isinstance(data_items[1], int):
            # eg [1:3, 0]
            if not isinstance(value, self.obj.ITERABLE_1D):
                value = [value] * (data_items[0].stop - data_items[0].start)
            try:
                value = value[list(self.obj.index)]
            except TypeError:
                pass
            data[
                data_items[0].start
                + step * data_items[1] : data_items[0].stop
                + data_items[1] * step
            ] = value

        if isinstance(data_items[0], int) and isinstance(data_items[1], slice):
            # eg .iloc[0, 1:3]
            start = data_items[0] + step * data_items[1].start
            stop = data_items[0] + step * data_items[1].stop
            if not isinstance(value, self.obj.ITERABLE_1D):
                value = [value] * (data_items[1].stop - data_items[1].start)
            try:
                value = value[list(self.obj.columns)]
            except TypeError:
                pass
            for i, val in zip(range(start, stop, step), value):
                data[i] = val
        if isinstance(data_items[0], int) and isinstance(
            data_items[1], self.obj.ITERABLE_1D
        ):
            # eg .iloc[0, [1, 2, 3]]
            if not isinstance(value, self.obj.ITERABLE_1D):
                value = [value] * (len(data_items[1]))
            for i, val in zip(items[1], value):
                data[data_items[0] + step * i] = val

        if isinstance(data_items[0], self.obj.ITERABLE_1D) and isinstance(
            data_items[1], int
        ):
            # eg .iloc[[1, 2, 3], 0]
            if not isinstance(value, self.obj.ITERABLE_1D):
                value = [value] * (len(data_items[0]))
            for i, val in zip(data_items[0], value):
                data[i + step * data_items[1]] = val

        #####################
        # Sets a 2D section
        #####################
        # warning: everything below is very messy.

        if isinstance(data_items[0], slice) and isinstance(data_items[1], slice):
            # e.g. .iloc[1:3, :]
            # there is almost certainly a better way to do this
            k = 0

            # convert the value to a flat list for assignment
            if isinstance(value, self.obj.ITERABLE_1D):
                value = list(itertools.chain.from_iterable(value))
            else:
                value = [value] * (
                    (data_items[1].stop - data_items[1].start)
                    * (data_items[0].stop - data_items[0].start)
                )

            for i in range(data_items[0].start, data_items[0].stop):
                for j in range(data_items[1].start, data_items[1].stop):
                    data[i + j * step] = value[k]
                    k += 1

        # handle a 2d boolean key
        if is_2d_bool(data_items[0]):
            try:
                data_items[0] = data_items[0].values
            except AttributeError:
                pass

            for i, row in enumerate(data_items[0]):
                for j, col in enumerate(row):
                    if col:
                        self.obj.data[
                            self.obj.bound_int_to_df(i, axis=0)
                            + self.obj.bound_int_to_df(j, axis=1) * self.obj.step
                        ] = value

        elif isinstance(data_items[0], self.obj.ITERABLE_1D) and isinstance(
            data_items[1], slice
        ):
            # e.g. .iloc[[1, 2], :]
            # there is almost certainly a better way to do this
            k = 0
            if isinstance(value, self.obj.ITERABLE_1D):
                value = list(itertools.chain.from_iterable(value))
            else:
                value = [value] * (
                    (data_items[1].stop - data_items[1].start) * len(data_items[0])
                )

            for i in data_items[0]:
                for j in range(data_items[1].start, data_items[1].stop):
                    data[i + j * step] = value[k]
                    k += 1
        if isinstance(data_items[0], slice) and isinstance(
            data_items[1], self.obj.ITERABLE_1D
        ):
            # e.g. .iloc[:, [1,2]
            k = 0
            if isinstance(value, self.obj.ITERABLE_1D):
                value = list(itertools.chain.from_iterable(value))
            else:
                value = [value] * (
                    (data_items[0].stop - data_items[0].start) * len(data_items[1])
                )

            for i in range(data_items[0].start, data_items[0].stop):
                for j in data_items[1]:
                    data[i + j * step] = value[k]
                    k += 1
        if isinstance(data_items[0], self.obj.ITERABLE_1D) and isinstance(
            data_items[1], self.obj.ITERABLE_1D
        ):
            # e.g. .iloc[:, [1,2]
            k = 0
            if isinstance(value, self.obj.ITERABLE_1D):
                value = list(itertools.chain.from_iterable(value))
            else:
                value = [value] * (len(data_items[0]) * len(data_items[1]))
            for i in data_items[0]:
                for j in data_items[1]:
                    data[i + j * step] = value[k]
                    k += 1


class ILocSer:
    """
    ILoc indexer for Series
    """

    ITERABLE_1D = (list, set, tuple)

    def __init__(self, obj):
        """
        Initializes the indexer
        :param obj: Series
        """
        self.obj = obj

    def __getitem__(self, item):
        """
        Setitem for Series based on index number
        """
        if isinstance(item, slice):
            item = slice(
                item.start if item.start is not None else 0,
                item.stop if item.stop is not None else len(self.obj),
            )
            view = self.obj.bound_slice(item)
            index = self.obj.index[item]
            return self.obj.from_data(self.obj.data, index, self.obj.name, view)

        if isinstance(item, self.ITERABLE_1D + (self.obj.__class__,)):
            if is_bool(item):
                item = [i for i, val in enumerate(item) if val]

            index = self.obj.index
            index = [index[i] if i is not None else None for i in item]
            data = self.obj.values
            data = [data[i] if i is not None else nan for i in item]
            view = slice(0, len(index), 1)
            return self.obj.from_data(data, index, self.obj.name, view)
        return self.obj.values[item]

    def __setitem__(self, item, value):
        """
        Setitem for Series based on index number
        """
        # convert to bool, or bound
        if isinstance(item, self.ITERABLE_1D + (self.obj.__class__,)):
            # if it's a boolean
            if is_bool(item):
                item = [i for i, val in enumerate(item) if val]
            data_item = self.obj.bound_iterable(item)
            if not isinstance(value, self.ITERABLE_1D + (self.obj.__class__,)):
                value = [value] * len(self.obj)

            for i, val in zip(data_item, value):
                self.obj.data[i] = val

        elif isinstance(item, slice):
            item = slice(
                item.start if item.start is not None else 0,
                item.stop if item.stop is not None else len(self.obj),
            )
            data_item = self.obj.bound_slice(item)
            if not isinstance(value, self.ITERABLE_1D + (self.obj.__class__,)):
                value = [value] * ((data_item.stop - data_item.start) // data_item.step)

            self.obj.data[data_item] = value

        else:
            # check the bounds
            if item >= len(self.obj):
                raise IndexError(
                    "You requested index %s but series is only %s items."
                    % (item, len(self.obj))
                )
            data_item = self.obj.bound_int(item)
            self.obj.data[data_item] = value


class LocSer:
    """
    Loc indexer for Series
    """

    ITERABLE_1D = (list, set, tuple)

    def __init__(self, obj):
        """
        Initializes the indexer
        :param obj: Series
        """
        self.obj = obj

    def __setitem__(self, items, value, what=None):
        """
        Setitem for Series based on index names
        """
        iloc_items = self.obj.index_of(items)

        # if index_of returned none, create it
        if iloc_items is None:
            self.obj.extend(items, num=1)
            self.__setitem__(items, value)
        else:
            self.obj.iloc.__setitem__(iloc_items, value)

    def __getitem__(self, items):
        """
        Getitem for Series based on index names
        """
        if is_bool(items):
            return self.obj.iloc[items]
        iloc_items = self.obj.index_of(items)
        if iloc_items is None:
            raise KeyError("%s not found in index." % items)
        ser = self.obj.iloc[iloc_items]
        if isinstance(ser, self.obj.__class__):
            ser.index = tuple(items)
        return ser


class LocDF:
    """
    Loc indexer for DataFrames
    """

    ITERABLE_1D = (list, set, tuple)

    def __init__(self, obj):
        """
        Initializes the indexer
        :param obj: DataFrame
        """
        self.obj = obj

    def __getitem__(self, items):
        """
        Getitem for DataFrames based on index names
        """
        if isinstance(items, tuple):
            if is_2d_bool(items[0]):
                return self.obj.iloc[items]
            # items arrive as slice and series
            iloc_items = tuple(
                self.obj.index_of(item, axis=i) for (i, item) in enumerate(items)
            )
        else:
            if is_2d_bool(items):
                return self.obj.iloc[items]
            iloc_items = self.obj.index_of(items)
        # can't use None in iloc_items. Fails with a series
        if any(elem is None for elem in iloc_items):
            raise KeyError(
                "One or more items not found. Index: %s, Column: %s"
                % (items[0], items[1])
            )
        return self.obj.iloc[iloc_items]

    def __setitem__(self, items, value):
        """
        Setitem for Series based on index names
        """
        # if it's a dataframe, send straight to iloc. It's a boolean key
        if is_2d_bool(items):
            self.obj.iloc.__setitem__(items, value)
            return

        if isinstance(items, tuple):
            if len(items) > 1 and is_2d_bool(items[1]):
                self.obj.iloc.__setitem__(items[1], value)
                return
            iloc_items = tuple(
                self.obj.index_of(item, axis=i) for (i, item) in enumerate(items)
            )
        else:
            iloc_items = (self.obj.index_of(items),)

        # if the index isn't found, add an empty row/column and call it again
        if iloc_items[0] is None:
            # adding a row will break the view. Make a copy.
            self.obj.drop()
            self.obj.add_empty_series(items[0], axis=0)
            self.__setitem__(items, value)
        elif len(items) > 1 and iloc_items[1] is None:
            self.obj.add_empty_series(items[1], axis=1)
            self.__setitem__(items, value)
        else:
            self.obj.iloc.__setitem__(iloc_items, value)
